#define MAXJOBS 16
#define JOBNAMEMAXCHARS 8

typedef unsigned long (*FunctionPointer) (int id);

struct task {
  FunctionPointer Function;   // Entry Point of Function
  char name[JOBNAMEMAXCHARS + 1];   // Name
  unsigned int id;            // Task ID
  unsigned long next;         // Time in ms from poweron to run next
  bool used;                  // true if in use
  bool active;                // true if active
};

task tasklist[MAXJOBS];
int tasknum = 0;

void InitTasklist() {
  int i;
  for (i = 0; i < MAXJOBS; i++) {
    tasklist[i].active = false;
    tasklist[i].used = false;
  }
}

// Add a task to Tasklist
// A Task is added to tasklist with
// AddTask(function, "Name8chr", (int) parameter, (int) parameter1, (int) parameter2);
// A task may look into the task table for its parameters
// The tasks itselves will be calles periodically by RunTasks:
// retval=(unsigned long) Function((int) index);
// index is the index into tasklist[] so that the function may lookup its id or parameters
// the return value is the waiting-time untill the next call in ms.

unsigned int AddTask(FunctionPointer p, char *taskname) {
  unsigned int i, j, slot = 0xffff;
  static unsigned int tasknextid = 1;
  if (tasknum < MAXJOBS ) {
    for (i = 0; i < MAXJOBS; i++) {
      if (tasklist[i].used == false) {
        slot = i;
        break;
      }
    }
    if (slot != 0xffff) {  // We have found a free slot for the task
      tasklist[slot].Function = p;
      for (i = 0; i <= JOBNAMEMAXCHARS; i++) {
        tasklist[slot].name[i] = '\0';
      }
      for (i = 0; i < JOBNAMEMAXCHARS && taskname[i] > 0; i++) {
        tasklist[slot].name[i] = taskname[i];
      }
      tasklist[slot].next = 0;
      tasklist[slot].used = true;
      tasklist[slot].active = true;
      tasklist[slot].id = tasknextid;
      tasknum++;
      i = tasknextid;
      do {  // Find next free TaskId. As we use an unsigned int it should never overflow on a Atmega but nobody knows so avoid collisions.
        tasknextid++; // First try is increment by one
        j = 0;
        for (slot = 0; slot < MAXJOBS; slot++) {
          if (tasklist[slot].used == true) {
            if (tasklist[slot].id == tasknextid) {
              j = 1;
            }
          }
        }
      }
      while (j == 1);
      return (i);
    }
  }
  return (0);
}

bool DelTask(unsigned int id) {
  int i, j = false;
  for (i = 0; i < MAXJOBS; i++) {
    if (tasklist[i].id == id) {
      tasklist[i].used = false;
      tasklist[i].active = false;
      tasknum--;
      j = true;
    }
  }
  return (j);
}

bool DeactivateTask(unsigned int id) {
  int i, j = false;
  for (i = 0; i < MAXJOBS; i++) {
    if (tasklist[i].id == id) {
      tasklist[i].active = false;
      tasknum--;
      j = true;
    }
  }
  return (j);
}

bool ActivateTask(unsigned int id) {
  int i, j = false;
  for (i = 0; i < MAXJOBS; i++) {
    if (tasklist[i].id == id) {
      tasklist[i].active = true;
      tasknum--;
      j = true;
    }
  }
  return (j);
}

void RunTasks() {
  int i;
  unsigned long now, retval;
  now = millis();
  for (i = 0; i < MAXJOBS; i++) {
    if (tasklist[i].active == true) {
      if (tasklist[i].next <= now) {
        retval = tasklist[i].Function(i);
        if (retval > 0) {
          tasklist[i].next = now + retval;
        }
      }
    }
  }
}
